#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from scipy import signal


def tophat2d(radius, width=10):
    """
    convolution kernel is a Mexican Hat revolved in x-plane

    Parameters
    ----------
    radius : scalar,
        peak position along the radius
    width : scalar,
        width of peak in pixels

    Returns
    -------
    ndarray
        kernel to convolve the signal
    """
    N = np.int(np.round(radius) + 3 * np.round(width) + 1)
    x = np.arange(-N, N)
    x, y = np.meshgrid(x, x)
    t = np.sqrt(x**2 + y**2) - radius
    a = 1. / np.sqrt(2 * np.pi) / width**3
    w = a * (1 - (t / width)**2) * np.exp(-t**2 / width**2 / 2.)
    return w


def cwt2d(image, domain=None, width=1, log=False):
    """
    continuous wavelets transform for finding rings in a SAXS calibration image

    Parameters
    ----------
    image: ndarray
        SAXS Calibration image
    domain: list
        [min, max] search region
    width: scalar
        width of the ring in pixels
    log: bool
        Use log of the image, default is False 
  
    Returns
    ------- 
    ndarray
        center of the detected ring
    """
    nrow, ncol = image.shape
    if domain is None:
        rmin = 0
        rmax = min(nrow, ncol)
    else:
        rmin = domain[0]
        rmax = domain[1]
    maxval = 0
    center = np.array([0, 0], dtype=np.int)
    radius = -1
    # if log is true
    if log: 
        sig = np.log(image+4).copy()
    else: sig = image.copy()

    for r in range(rmin, rmax):
        w = tophat2d(r, width)
        im2 = signal.fftconvolve(sig, w, 'same')
        if im2.max() > maxval:
            maxval = im2.max()
            center = np.unravel_index(im2.argmax(), image.shape)
            radius = r
    return center, radius



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import numpy as np
from pyFAI.geometry import Geometry
from pyFAI.ext import splitBBox
from typing import Tuple
from pyFAI import units

# Limitations of this module:
# This module incompletely supports tilt/rotation. The alpha/phi angles are calculated correctly from the geometry,
# but its unclear how these should map to the reflection geometry -> tilt/rotation are unsupported in reflection.
#
# The transmission -> reflection mapping is done by subtracting alpha_i from the transmission alphas, effectively making
# the horizon at alpha=0. This mapping does not correct the solid angle normalization.
#
# The pyFAI integrators also have strange windowing behavior; if part of the image is cutoff, you need to transform it
# into a reduced coordinate system (and back afterwards).


def q_from_angles(phi, alpha, wavelength):
    r = 4 * np.pi / wavelength
    qx = r * np.sin(phi/2) * np.cos(alpha/2)
    qy = r * np.sin(alpha/2)
    qz = r * np.cos(alpha/2) * np.cos(alpha/2) - 1

    return np.array([qx, qy, qz])


def alpha(x, y, z):
    return np.arctan2(y, np.sqrt(x ** 2 + z ** 2))


def phi(x, y, z):
    return np.arctan2(x, np.sqrt(z ** 2))


def test_show(data, log=True):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    if log:
        image = np.log(data + 3)
    else:
        image = data
    im = plt.imshow(image)
    fig.colorbar(im, ax=fig.axes)
    plt.show()


def remesh(image: np.ndarray,
           geometry: Geometry,
           reflection: bool,
           alphai: float,
           bins: Tuple[int, int] = None,
           q_h_range: Tuple[float, float] = None,
           q_v_range: Tuple[float, float] = None,
           out_range=None,
           res=None,
           coord_sys='qp_qz'):
    """
    Redraw the GI Image in (qp, qz) coordinates.

    Parameters:
    -----------
    image: ndarray, 
        detector image 
    geometry: pyFAI Geometry
        PONI, Pixel Size, Sample Det dist etc
    alphai: scalar, deg
        angle of incedence
    out_range: list, optional
        [[left, right],[lower, upper]] for the output image
    res: list, optional
        resolution of the output image
    coord_sys: str
        'qp_qz', 'qy_qz' or 'theta_alpha' 
    Returns
    -------
    qimage: ndarray
        remeshed/warped image
    xcrd: ndarray
        x-coords of warped image
    ycrd: ndarray
        y-coords of warped image
    """
    assert image.shape == geometry.detector.shape
    x = np.arange(image.shape[1])
    y = np.arange(image.shape[0])
    px_x, px_y = np.meshgrid(x, y)
    r_z, r_y, r_x = geometry.calc_pos_zyx(d1=px_y, d2=px_x)

    if reflection:
        alphas = alpha(r_x, r_y, r_z) - alphai
        phis = phi(r_x, r_y, r_z)

        k_i = k_f = 2 * np.pi / geometry.wavelength * 1e-10

        q_x = k_f * np.cos(alphas) * np.cos(phis) - k_i * np.cos(alphai)
        q_y = k_f * np.cos(alphas) * np.sin(phis)
        q_z = -(k_f * np.sin(alphas) + k_i * np.sin(alphai))
        q_r = np.sqrt(q_x ** 2 + q_y ** 2) * np.sign(q_y)
        q_h = q_r
        q_v = -q_z
    else:
        alphas = alpha(r_x, r_y, r_z)
        phis = phi(r_x, r_y, r_z)

        q_x, q_y, q_z = q_from_angles(phis, alphas, geometry.wavelength) * 1e-10
        q_v = -q_y  # -q_y
        q_h = q_x

    if bins is None: bins = tuple(image.shape)
    if q_h_range is None: q_h_range = (q_h.min(), q_h.max())
    if q_v_range is None: q_v_range = (q_v.min(), q_v.max())

    I, q_h, q_v, _, _ = splitBBox.histoBBox2d(weights=image,
                                              pos0=q_h,
                                              delta_pos0=np.ones_like(image) * (q_h_range[1] - q_h_range[0]) / bins[0],
                                              pos1=q_v,
                                              delta_pos1=np.ones_like(image) * (q_v_range[1] - q_v_range[0]) / bins[1],
                                              bins=bins,
                                              pos0Range=q_h_range,
                                              pos1Range=q_v_range,
                                              dummy=None,
                                              delta_dummy=None,
                                              allow_pos0_neg=True,
                                              # mask=mask,
                                              # dark=dark,
                                              # flat=flat,
                                              # solidangle=solidangle,
                                              # polarization=polarization,
                                              # normalization_factor=normalization_factor,
                                              # chiDiscAtPi=self.chiDiscAtPi,
                                              # empty=dummy if dummy is not None else self._empty
                                              )
    q_h, q_v = np.meshgrid(q_h, q_v)
    return I, q_h, q_v


if __name__ == '__main__':
    from pyFAI import detectors
    import fabio

    filename = '/home/rp/data/YL1031/AGB_5S_USE_2_2m.edf'

    fit2d = {'centerX': 237.5,
             'centerY': 1678 - 30.500000000000004,
             'directDist': 283.26979219190343,
             'pixelX': 171.99999999999997,
             'pixelY': 171.99999999999997,
             'splineFile': None,
             'tilt': 0.0,
             'tiltPlanRotation': 0.0}

    det = detectors.Pilatus2M
    g = Geometry(wavelength=0.123984E-09)
    g.setFit2D(**fit2d)
    g.detector = det()

    data = fabio.open(filename).data
    ai = np.deg2rad(0.14)

    test_show(data)
    from pyFAI import calibrant

    test_show(calibrant.ALL_CALIBRANTS['AgBh'].fake_calibration_image(g))

    # default qp-qz, same size as detector
    I, q_h, q_v = remesh(data, g, False, 0)  # img, x, y =
    test_show(I)



#! /usr/bin/env python

from sasmodels.core import load_model
from sasmodels.direct_model import call_kernel
from astropy.modeling import Parameter
from astropy.modeling import Fittable1DModel

def XicamSASModel(name, params):
    """
    XicamModel wraps sasmdoels from sasview (https://github.com/SasView/sasmodles),
    in a fittable 1-D plugin for Xi-cam. This object can be passed to astropy for
    data fitting.
    
    Parameters: 
        name (str) : name of the sasmoels (http://www.sasview.org/sasmodels/index.html)
        params (dict): key : value pairs of fittable parameters
    Returns:
        func (XicamFittable) : function that takes q-values and returns intensity values

"""
    inputs = [p.name for p in params]
    model_name = name.lower().replace(' ','_')
    m = load_model(model_name)

    # evaluate callback
    def saxs_curve(q, *args):
        kernel = m.make_kernel([q])
        p_fit = dict(zip(inputs, args))
        return call_kernel(kernel, p_fit)

    # create an astropy fittable model
    names = {
        'name': 'SASFittable_'+name+'Model',
        'inputs': inputs,
        'outputs': ['I'],
        'evaluate': staticmethod(saxs_curve)
    }

    names['fixed'] = dict((p.name, p.fixed) for p in params)
    names['bounds'] = dict((p.name, p.bounds) for p in params)
    create_param = lambda p : Parameter(p.name, default = p.value)
    _parameters_ = dict((p.name, create_param(p)) for p in params)
    names.update(_parameters_)
    return type('SASModelFittable', (Fittable1DModel,), names)()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import numpy as np
from .warp_image import warp_image, pixel2q, x2angles
try:
    import cWarpImage
    remesh_fcn = cWarpImage.warp_image
except ImportError:
    remesh_fcn = warp_image


def remesh(image,
           geometry,
           alphai,
           out_range=None,
           res=None,
           coord_sys='qp_qz'):
    """
    Redraw the GI Image in (qp, qz) coordinates.

    Parameters:
    -----------
    image: ndarray, 
        detector image 
    geometry: pyFAI Geometry
        PONI, Pixel Size, Sample Det dist etc
    alphai: scalar, deg
        angle of incedence
    out_range: list, optional
        [[left, right],[lower, upper]] for the output image
    res: list, optional
        resolution of the output image
    coord_sys: str
        'qp_qz', 'qy_qz' or 'theta_alpha' 
    Returns
    -------
    qimage: ndarray
        remeshed/warped image
    xcrd: ndarray
        x-coords of warped image
    ycrd: ndarray
        y-coords of warped image
    """
    rows, cols = image.shape
    center = np.array([geometry.get_poni2(), geometry.get_poni1()])
    pixel = [geometry.get_pixel1(), geometry.get_pixel2()]
    dist = geometry.get_dist()
    # convert wavelen to nanomters
    wavelen = geometry.get_wavelength() * 1.0E+09
    k0 = 2 * np.pi / wavelen

    if coord_sys is None:
        coord_sys = 'qp_qz'

    if res is None:
        res = [rows, cols]
    else:
        if not len(res) == 2:
            sys.stderr.write('resolution should a sequence of two integers')
            exit(1)

    if out_range is None:
        # coordinates for detector corners
        coord = np.multiply([[0, cols], [0, rows]], pixel) - center[:,np.newaxis]
        if coord_sys == 'qp_qz' or coord_sys == 'qy_qz':
            qx, qy, qz = pixel2q(coord[0, :], coord[1, :], dist, alphai, k0)
            qp = np.sign(qy) * np.sqrt(qx**2 + qy**2)
            qz = np.linspace(qz[0], qz[1], res[0])
            qp = np.linspace(qp[0], qp[1], res[1])
            ycrd = qz
            if coord_sys == 'qp_qz':
                xcrd, ycrd = np.meshgrid(qp, qz)
                qp, qz = xcrd, ycrd
            else:
                xcrd, ycrd = np.meshgrid(qy, qz)
                qp, qz = np.meshgrid(qp, qz)
                

        if coord_sys == 'theta_alpha':
            theta, alpha = x2angles(coord[0,:], coord[1,:], dist)

            xcrd = np.linspace(theta[0], theta[1], res[1])
            ycrd = np.linspace(alpha[0], alpha[1], res[0])
            xcrd, ycrd = np.meshgrid(xcrd, ycrd)

            qx = k0 * (np.cos(alpha) * np.cos(theta) - np.cos(alphai))
            qy = k0 * (np.cos(alpha) * np.sin(theta))
            qz = k0 * (np.sin(alpha) + np.sin(alphai))
            qp = np.sign(qy) * np.sqrt(qx**2 + qy**2)
            qz = np.linspace(qz[0], qz[1], res[0])
            qp = np.linspace(qp[0], qp[1], res[1])
            qp, qz = np.meshgrid(qp, qz)
            
    else:
        if coord_sys == 'qp_qz':
            qp = np.linspace(out_range[0][0], out_range[0][1], res[1])
            qz = np.linspace(out_range[1][0], out_range[1][1], res[0])
            qp, qz = np.meshgrid(qp, qz)
            xcrd, ycrd = qp, qz

        elif coord_sys == 'qy_qz':
            qz = np.array(out_range[1])
            qy = np.array(out_range[0])
            xcrd, ycrd = np.meshgrid(qy, qz)
            sin_al = qz/k0 - np.sin(alphai)
            cos_al = np.sqrt(1. - sin_al**2)
            sin_th = qy/k0/cos_al
            cos_th = np.sqrt(1. - sin_th**2)
            qx = k0 * (cos_al * cos_th - np.cos(alphai))
            qp = np.sign(qy) * np.sqrt(qx**2 + qy**2)
            qp = np.linspace(qp[0], qp[1], res[1])
            qz = np.linspace(qz[0], qz[1], res[0])
            qp, qz = np.meshgrid(qp, qz)
            

        elif coord_sys == 'theta_alpha':
            xcrd = np.array(out_range[0])
            ycrd = np.array(out_range[1])
            qx = k0 * (np.cos(ycrd) * np.cos(xcrd) - np.cos(alphai))
            qy = k0 * np.cos(ycrd) * np.sin(xcrd)
            qz = k0 * (np.sin(ycrd) + np.sin(alphai))
            qp = np.sign(qy) * np.sqrt(qx**2 + qy**2)
            qp = np.linspace(qp[0], qp[1], res[1])
            qz = np.linspace(qz[0], qz[1], res[0])
            qp, qz = np.meshgrid(qp, qz)
            xcrd = np.linspace(xcrd[0], xcrd[1], res[1])
            ycrd = np.linspace(ycrd[0], ycrd[1], res[0])
            xcrd, ycrd = np.meshgrid(xcrd, ycrd)
            
    qimage = remesh_fcn(image, qp, qz, pixel, center, alphai, k0, dist, 0)
    return qimage, xcrd, ycrd



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
import numpy as np

def warp_image(image, qp, qz, pixel, center, alphai, k0, dist, method):
    # find reverse map for every pair or qp,qz
    sin_al = qz / k0 - np.sin(alphai)
    cos_al = np.sqrt(1 - sin_al**2)
    with np.errstate(divide='ignore'):
        tan_al = sin_al / cos_al

    cos_ai = np.cos(alphai)
    with np.errstate(divide='ignore'):
        cos_th = (cos_al**2 + cos_ai**2 - (qp / k0)**2) / (2 * cos_al * cos_ai)
    with np.errstate(invalid='ignore'):
        sin_th = np.sign(qp) * np.sqrt(1 - cos_th**2)
    with np.errstate(divide='ignore', invalid='ignore'):
        tan_th = sin_th / cos_th

    map_x = ((tan_th * dist + center[0]) / pixel[0])
    with np.errstate(divide='ignore'):
        map_y = ((tan_al * dist / cos_th + center[1]) / pixel[1])

    # get forbidden locations
    mask = np.isnan(map_x) | np.isnan(map_y)
    rows,cols = image.shape

    # remove nans
    map_x[mask] = 0
    map_y[mask] = 0

    # mask out of range values
    mask[(map_x < 0)|(map_x > cols-1)] = True
    mask[(map_y < 0)|(map_y > rows-1)] = True

    # remove out of range values
    map_x[mask] = 0
    map_y[mask] = 0

    # flatten and convert to list of integers
    map_x = map_x.astype(int).ravel().tolist()
    map_y = map_y.astype(int).ravel().tolist()
    q_image = np.fromiter([image[i,j] for i,j in zip(map_y,map_x)], np.float).reshape(qp.shape)
    q_image[mask] = 0.
    return q_image


def x2angles(x, y, d):
    s1 = np.sqrt(d**2 + x**2)
    theta = np.arctan2(x, d)
    alpha = np.arctan2(y, s1)
    return theta, alpha


def pixel2q(x, y, d, alphai, k0):
    s1 = np.sqrt(x**2 + d**2)
    s2 = np.sqrt(s1**2 + y**2)
    sin_th = x / s1
    cos_th = d / s1
    sin_al = y / s2
    cos_al = s1 / s2
    qx = k0 * (cos_al * cos_th - np.cos(alphai))
    qy = k0 * cos_al * sin_th
    qz = k0 * (sin_al + np.sin(alphai))
    return qx, qy, qz



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from astropy.modeling import Fittable1DModel, fitting
from collections import OrderedDict

def fit_sasmodel(model, q, data, method='LevMarLSQFitter', **kwargs):
    """
    Fit an XicamSASModle using astropy.ftting routines.

    Parameters
    ----------
    model : XicamSASModel
        Model generated by the factroy method
    q     : np.ndarray
        q-range for the fitting
    data  : np.ndarray
        SAXS spectra over input q-range
    method: str
        one of [LinearLSQFitter, LevMarLSQFitter, SLSQPLSQFitter]
    kwargs: dict
        astropy fitting related key: arguments

    Returns
    -------
        Fitted parameters, values
    """

    # supported fitters
    fitters = OrderedDict({
        'LinearLSQFitter': fitting.LinearLSQFitter,
        'LevMarLSQFitter': fitting.LevMarLSQFitter,
        'SLSQPLSQFitter' : fitting.SLSQPLSQFitter
        })

    # sanity checks
    if not isinstance(model, Fittable1DModel):
        raise TypeError('Input model must be astropy Fittable1DModel or subtype')

    if not np.shape(q) == np.shape(data):
        raise ValueError('q and I must be of same shape')

    if not method in fitters:
        raise ValueError('Unsupported fitting methods')

    # give it your best shot. maybe!
    fit_t = fitters[method]()
    g = fit_t(model, q, data, **kwargs)
    return g.parameters, g(q)



# -*- coding: utf-8 -*-

from .remesh_bbox import remesh
from .cwt import cwt2d
from .factory import XicamSASModel
from .fit_sasmodel import fit_sasmodel


# load sasmodels
from .loader import load_models
try:
    models = load_models()
except:
    print('Warning: failed to load SASMODLES.')
    print('Please consider submitting a but-report') 




#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import yaml
from collections import OrderedDict

# Fitting related
_fixed = {'name': 'Fixed', 'value': True, 'type': 'bool'}
_bounds = [{'name': 'Lower', 'value': '-\u221E'}, {'name': 'Upper', 'value': '\u221E'}]
_bounded = {'name': 'Bounded', 'value': False, 'type': 'bool',
            'children': _bounds, 'visible': False, 'enabled': False}


class XicamParameter(yaml.YAMLObject):
    yaml_tag = u'!YMLParameter'

    def __init__(self, name, description, value, units, **kwags):
        self.name = name
        self.description = description
        self.value = value
        self.units = units
        self.fixed = True
        self.bounds = [None, None]
    def __repr__(self):
        return "%s (%r, value=%r, units=%r)" % \
               (self.__class__.__name__, self.name, self.value, self.units)

    @classmethod
    def from_yaml(cls, loader, node):
        opts = loader.construct_mapping(node)
        return cls(**opts)


def load_models():
    path,_= os.path.split(os.path.realpath(__file__))
    config = os.path.join(path, 'config.yml')
    if not os.path.isfile(config):
        raise ImportWarning('Unable to load config file')
        return None

    with open(config) as fp:
        yml = yaml.load(fp)

    model_tree = OrderedDict()
    for key, val in yml.items():
        models = OrderedDict()
        for name, params in val.items():
            _params = [p['param'] for p in params]
            models[name] = {'params': _params}
        model_tree[key] = models
    return model_tree




from setuptools import setup

if __name__ == '__main__':
    setup(name='camsaxs',
        version='1.0.1',
        description='Xi-cam.SAXS companion functions',
        author='Dinesh Kumar',
        author_email='dkumar@lbl.gov',
        url = "http://github.com/lbl-camera/CamSAXS",
        install_requires = ['numpy', 'scipy', 'astropy', 'pyFAI', 'sasmodels', 'pyyaml'],
        packages = ['camsaxs'],
        data_files=[('camsaxs', ['camsaxs/config.yml'])]
    )
